# -*- coding: utf-8 -*-
# (c) 2023 Brian Scholer (@briantist)

import json

import typing as t
from typing import ValuesView

from semver import VersionInfo
from collections import UserDict
from dataclasses import dataclass
from functools import total_ordering, cached_property
from datetime import datetime

from artifactory import ArtifactoryPath


@total_ordering
@dataclass(eq=False, order=False)
class CollectionData:
    collection_info: dict
    created_datetime: datetime
    modified_datetime: datetime
    namespace: str
    name: str
    filename: str
    sha256: str
    size: int
    mime_type: str
    version: str

    @classmethod
    def from_artifactory_path(cls, *, path: ArtifactoryPath, properties: dict, stat: dict):
        collection_info = json.loads(properties['collection_info'][0])

        return cls(
            collection_info=collection_info,
            created_datetime=stat.ctime,
            modified_datetime=stat.mtime,
            namespace=properties['namespace'][0],
            name=properties['name'][0],
            filename=path.name,
            sha256=stat.sha256,
            size=stat.size,
            mime_type=stat.mime_type,
            version=properties['version'][0],
        )

    def __eq__(self, other) -> bool:
        # All of the comparable information is in the collection
        # tarball itself, so this should be a nice fast way of
        # comparing equality, without computing the cached_properties.
        try:
            return self.sha256 == other.sha256
        except AttributeError:
            # Allow for comparing directly with VersionInfo objects.
            if isinstance(other, VersionInfo):
                return self.semver == other
            return NotImplemented

    def __lt__(self, other) -> bool:
        this_is_prerelease = self.is_prerelease
        this_version = self.semver

        if isinstance(other, VersionInfo):
            other_is_prerelease = other.prerelease is not None
            other_version = other
        elif isinstance(other, CollectionData):
            other_is_prerelease = other.is_prerelease
            other_version = other.semver
            if self.name != other.name or self.namespace != other.namespace:
                return False
        else:
            return NotImplemented

        # Both objects are prerelease versions, or both are not, so compare normally.
        if this_is_prerelease == other_is_prerelease:
            return this_version < other_version

        # One of the objects is not a prerelease and the other is.
        # If self is a prerelease, then it is necessarily < the other.
        return this_is_prerelease

    @cached_property
    def fqcn(self) -> str:
        return f"{self.namespace}.{self.name}"

    @cached_property
    def created(self) -> str:
        return self.created_datetime.isoformat()

    @cached_property
    def modified(self) -> str:
        return self.modified_datetime.isoformat()

    @cached_property
    def semver(self) -> VersionInfo:
        return VersionInfo.parse(self.version)

    @cached_property
    def is_prerelease(self) -> bool:
        return self.semver.prerelease is not None


class CollectionGroup(UserDict):
    """
    A dict-like object that represents one or more
    versions of a single collection. The keys are
    VersionInfo objects and the values are
    CollectionData objects.
    """
    latest: CollectionData = None

    def __init__(self, *, namespace: str, name: str):
        self.namespace = namespace
        self.name = name
        super().__init__()

    @classmethod
    def from_collection(cls, collection: CollectionData):
        instance = cls(namespace=collection.namespace, name=collection.name)
        instance.add(collection)
        return instance

    @property
    def versions(self) -> t.Dict[VersionInfo, CollectionData]:
        return self.data

    @cached_property
    def fqcn(self) -> str:
        return f"{self.namespace}.{self.name}"

    def add(self, collection: CollectionData) -> None:
        self[collection.semver] = collection

    @staticmethod
    def _get_key(key: t.Union[str, VersionInfo], *, raises: bool = True) -> VersionInfo:
        if isinstance(key, VersionInfo):
            return key
        elif isinstance(key, str):
            return VersionInfo.parse(key)
        else:
            if raises:
                raise TypeError("Only valid semantic versions can be used as keys.")
            else:
                return key

    def __getitem__(self, key: t.Union[str, VersionInfo]) -> CollectionData:
        return super().__getitem__(self._get_key(key))

    def __setitem__(self, key: t.Union[str, VersionInfo], item: CollectionData) -> None:
        if not isinstance(item, CollectionData):
            raise TypeError("Values must be of type CollectionData.")

        if self.name != item.name or self.namespace != item.namespace:
            raise ValueError(f"Attempted to add collection '{item.namespace}.{item.name}' to group for '{self.namespace}.{self.name}'.")

        if self.latest is None or item > self.latest:
            self.latest = item

        return super().__setitem__(self._get_key(key), item)

    def __delitem__(self, key: t.Union[str, VersionInfo]) -> None:
        dkey = self._get_key(key)
        super().__delitem__(dkey)

        if self.latest == dkey:
            self.latest = max(self.values(), default=None)

    def __contains__(self, key: t.Union[str, VersionInfo]) -> bool:
        return super().__contains__(self._get_key(key, raises=False))

    # re-define for the type hints
    def values(self) -> ValuesView[CollectionData]:
        return super().values()


class CollectionCollection(UserDict):
    """
    A Dict[str, CollectionGroup] object where the keys are FQCNs.
    """
    @classmethod
    def from_collections(cls, collections: t.Iterable[CollectionData]):
        instance = cls()
        for collection in collections:
            instance.add(collection)
        return instance

    def add(self, collection: CollectionData) -> None:
        if collection.fqcn in self.data:
            self.data[collection.fqcn].add(collection)
        else:
            self.data[collection.fqcn] = CollectionGroup.from_collection(collection)

    # re-define for the type hints
    def values(self) -> ValuesView[CollectionGroup]:
        return super().values()



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from . import create_configured_app

if __name__ == '__main__':
    create_configured_app(run=True, parse_known_only=False, parse_allow_abbrev=True)



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from flask import abort, request, current_app, send_file

from . import bp as dl
from .. import constants as C
from ..utilities import authorize, _chunk_to_temp, upload_collection_from_hashed_tempfile
from ..upstream import ProxyUpstream


@dl.route('/<filename>')
def download(filename):
    artifact = authorize(request, current_app.config['ARTIFACTORY_PATH'] / filename)
    upstream = current_app.config['PROXY_UPSTREAM']
    # no_proxy = current_app.config['NO_PROXY_NAMESPACES']
    cache_minutes = current_app.config['CACHE_MINUTES']
    cache_read = current_app.config['CACHE_READ']
    cache_write = current_app.config['CACHE_WRITE']
    property_fallback = current_app.config.get('USE_PROPERTY_FALLBACK', False)

    try:
        stat = artifact.stat()
    except FileNotFoundError:
        # Although there's a naming convention for the collection tarballs, the name is not actually
        # at all significant for the download/install process; previous API calls gave the download
        # URL exactly. So we don't actually have namespace information to check for no-proxying.
        # TODO: consider whether we should filter by the naming convention, maybe configurable.
        if not upstream:  # or not (not no_proxy or namespace not in no_proxy):
            abort(C.HTTP_NOT_FOUND)

        proxy = ProxyUpstream(artifact, upstream, cache_read, cache_write, cache_minutes)

        with proxy.proxy_download(request) as resp, _chunk_to_temp(None, iterator=resp.iter_content, close=cache_write) as tmp:
            if not cache_write:
                return send_file(tmp.handle, as_attachment=True, download_name=filename, etag=False)

            upload_collection_from_hashed_tempfile(artifact, tmp, property_fallback=property_fallback)

        stat = artifact.stat()

    return send_file(artifact.open(), as_attachment=True, download_name=artifact.name, last_modified=stat.mtime, etag=False)



# -*- coding: utf-8 -*-
# (c) 2023 Brian Scholer (@briantist)

import pytest
import typing as t

from datetime import datetime, timezone
from functools import partial

from galactory.models import CollectionData


@pytest.fixture
def collection_data_factory() -> t.Callable[[], CollectionData]:
    values = dict(
        collection_info={},
        namespace='ns',
        name='name',
        created_datetime=datetime.now(timezone.utc),
        modified_datetime=datetime.now(timezone.utc),
        filename='fake-file',
        mime_type='fake-file',
        sha256='m-m-m-my-sha-',
        size=0,
        version='0.0.0',
    )
    return partial(CollectionData, **values)


@pytest.fixture
def collection_data(request, collection_data_factory) -> CollectionData:
    overrides = getattr(request, 'param', {})
    return collection_data_factory(**overrides)



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

CONTENT_TYPE = {'Content-Type': 'application/json'}
HTTP_OK = 200
HTTP_NOT_FOUND = 404
HTTP_INTERNAL_SERVER_ERROR = 500

QUERY_DOWNLOAD_UPSTREAM_URL = 'galactory_upstream_url'



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

import pytest
import os
import json

from unittest import mock
# pytest_plugins = ["docker_compose"]


@pytest.fixture(scope='session')
def disable_http():
    def _unexpected_request(self, method, url, *args, **kwargs):
        raise RuntimeError(f"Attempt to make {method} request to {self.scheme}://{self.host}{url}")

    with mock.patch('urllib3.connectionPool.urlopen', _unexpected_request):
        yield


@pytest.fixture(scope='session')
def fixture_finder():
    def _finder(*paths):
        here = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.join(here, 'fixtures', *paths)
        return fixture
    return _finder


@pytest.fixture(scope='session')
def fixture_loader(fixture_finder):
    class FixtureLoader:
        @staticmethod
        def _finder(relative_path, implied_extension=None):
            fixture = fixture_finder(relative_path)
            if implied_extension and not os.path.splitext(fixture)[1]:
                fixture = '.'.join([fixture, implied_extension])

            return fixture

        @classmethod
        def load_json(cls, relative_path):
            with open(cls._finder(relative_path, implied_extension='json'), 'r') as f:
                return json.load(f)

    return FixtureLoader



# -*- coding: utf-8 -*-
# (c) 2023 Brian Scholer (@briantist)

from itertools import product

from galactory.models import CollectionCollection


def test_collectioncollection_init_add(collection_data_factory):
    ci = collection_data_factory()

    cc = CollectionCollection()
    assert len(cc) == 0
    cc.add(collection_data_factory())
    assert len(cc) == 1
    assert len(cc[ci.fqcn]) == 1
    cc.add(collection_data_factory())
    assert len(cc) == 1 # same name
    assert len(cc[ci.fqcn]) == 1 # same version

    cnewver = collection_data_factory(version='9.8.7', sha256='Z')
    cc.add(cnewver)
    assert len(cc) == 1 # same name
    assert len(cc[ci.fqcn]) == 2
    assert cc[ci.fqcn].latest is cnewver


def test_collectioncollection_from_collections(collection_data_factory):
    namespaces = ['ns1', 'ns2']
    names = ['n1', 'n2']
    versions = ['1.2.3', '1.2.3-dev0', '9.8.7', '0.0.0']

    collections = [
        collection_data_factory(namespace=d[0], name=d[1], version=d[2], sha256=str(i))
        for i, d in enumerate(product(namespaces, names, versions))
    ]

    cc = CollectionCollection.from_collections(collections)

    assert len(cc) == len(namespaces) * len(names)

    for cg in cc.values():
        assert len(cg.versions) == len(versions)
        assert cg.latest.version == '9.8.7'



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from semver import VersionInfo
from flask import Response, jsonify, abort, url_for, request, current_app

from . import bp as v2
from ... import constants as C
from ...utilities import (
    discover_collections,
    authorize,
    _chunk_to_temp,
    upload_collection_from_hashed_tempfile,
    IncomingCollectionStream,
)
from ...upstream import ProxyUpstream
from ...models import CollectionCollection


@v2.route('/collections')
@v2.route('/collections/', endpoint='collections')
def collections():
    repository = authorize(request, current_app.config['ARTIFACTORY_PATH'])
    scheme = current_app.config.get('PREFERRED_URL_SCHEME')

    results = []
    colcol = CollectionCollection.from_collections(discover_collections(repo=repository))

    for colgroup in colcol.values():
        result = {
            'href': url_for(
                ".collection",
                namespace=colgroup.namespace,
                collection=colgroup.name,
                _external=True,
                _scheme=scheme,
            ),
            'name': colgroup.name,
            'namespace': {
                'name': colgroup.namespace,
            },
            'deprecated': False, # FIXME
            'created': colgroup.latest.created,
            'modified': colgroup.latest.modified,
            'versions_url': url_for(
                ".versions",
                namespace=colgroup.latest.namespace,
                collection=colgroup.latest.name,
                _external=True,
                _scheme=scheme,
            ),
            'latest_version': {
                'href': url_for(
                    ".version",
                    namespace=colgroup.latest.namespace,
                    collection=colgroup.latest.name,
                    version=colgroup.latest.version,
                    _external=True,
                    _scheme=scheme,
                ),
                "version": colgroup.latest.version,
            },
        }
        results.append(result)

    out = {
        'count': len(results),
        'results': results,
        'next': None,
        'next_link': None,
        'previous': None,
        'previous_link': None,
    }

    return out


@v2.route('/collections/<namespace>/<collection>')
@v2.route('/collections/<namespace>/<collection>/', endpoint='collection')
def collection(namespace, collection):
    repository = authorize(request, current_app.config['ARTIFACTORY_PATH'])
    upstream = current_app.config['PROXY_UPSTREAM']
    no_proxy = current_app.config['NO_PROXY_NAMESPACES']
    cache_minutes = current_app.config['CACHE_MINUTES']
    cache_read = current_app.config['CACHE_READ']
    cache_write = current_app.config['CACHE_WRITE']
    scheme = current_app.config.get('PREFERRED_URL_SCHEME')

    upstream_result = None
    if upstream and (not no_proxy or namespace not in no_proxy):
        proxy = ProxyUpstream(repository, upstream, cache_read, cache_write, cache_minutes)
        upstream_result = proxy.proxy(request)

    colcol = CollectionCollection.from_collections(discover_collections(repo=repository, namespace=namespace, name=collection))


    if not (colcol or upstream_result):
        abort(C.HTTP_NOT_FOUND)

    colgroup = None
    if colcol:
        if len(colcol) > 1:
            abort(C.HTTP_INTERNAL_SERVER_ERROR)
        colgroup = next(iter(colcol.values()))

    if upstream_result:
        if colgroup is None:
            return upstream_result
        else:
            try:
                upstream_version = VersionInfo.parse(upstream_result['latest_version']['version'])
            except (KeyError, ValueError):
                # TODO: warn?
                pass
            else:
                if colgroup.latest < upstream_version:
                    return upstream_result

    result = {
        'href': url_for(
            ".collection",
            namespace=colgroup.namespace,
            collection=colgroup.name,
            _external=True,
            _scheme=scheme,
        ),
        'name': colgroup.latest.name,
        'namespace': {
            'name': colgroup.latest.namespace,
        },
        'deprecated': False, # FIXME
        'created': colgroup.latest.created,
        'modified': colgroup.latest.modified,
        'versions_url': url_for(
            ".versions",
            namespace=colgroup.latest.namespace,
            collection=colgroup.latest.name,
            _external=True,
            _scheme=scheme,
        ),
        'latest_version': {
            'href': url_for(
                ".version",
                namespace=colgroup.latest.namespace,
                collection=colgroup.latest.name,
                version=colgroup.latest.version,
                _external=True,
                _scheme=scheme,
            ),
            "version": colgroup.latest.version,
        },
    }
    return result


@v2.route('/collections/<namespace>/<collection>/versions')
@v2.route('/collections/<namespace>/<collection>/versions/', endpoint='versions')
def versions(namespace, collection):
    results = []
    repository = authorize(request, current_app.config['ARTIFACTORY_PATH'])
    upstream = current_app.config['PROXY_UPSTREAM']
    no_proxy = current_app.config['NO_PROXY_NAMESPACES']
    cache_minutes = current_app.config['CACHE_MINUTES']
    cache_read = current_app.config['CACHE_READ']
    cache_write = current_app.config['CACHE_WRITE']
    scheme = current_app.config.get('PREFERRED_URL_SCHEME')

    upstream_result = None
    if upstream and (not no_proxy or namespace not in no_proxy):
        proxy = ProxyUpstream(repository, upstream, cache_read, cache_write, cache_minutes)
        upstream_result = proxy.proxy(request)

    collections = CollectionCollection.from_collections(discover_collections(repo=repository, namespace=namespace, name=collection))

    if not (collections or upstream_result):
        abort(C.HTTP_NOT_FOUND)

    if len(collections) > 1:
        abort(C.HTTP_INTERNAL_SERVER_ERROR)

    vers = set()
    for col in collections.values():
        for i in col.values():
            results.append(
                {
                    'href': url_for(
                        ".version",
                        namespace=i.namespace,
                        collection=i.name,
                        version=i.version,
                        _external=True,
                        _scheme=scheme,
                    ),
                    'version': i.version,
                }
            )
            vers.add(i.version)

    if upstream_result:
        for item in upstream_result['results']:
            if item['version'] not in vers:
                results.append(item)

    out = {
        'count': len(results),
        'next': None,
        'next_link': None,
        'previous': None,
        'previous_link': None,
        'results': results,
    }
    return out


@v2.route('/collections/<namespace>/<collection>/versions/<version>')
@v2.route('/collections/<namespace>/<collection>/versions/<version>/', endpoint='version')
def version(namespace, collection, version):
    repository = authorize(request, current_app.config['ARTIFACTORY_PATH'])
    upstream = current_app.config['PROXY_UPSTREAM']
    no_proxy = current_app.config['NO_PROXY_NAMESPACES']
    cache_minutes = current_app.config['CACHE_MINUTES']
    cache_read = current_app.config['CACHE_READ']
    cache_write = current_app.config['CACHE_WRITE']
    scheme = current_app.config.get('PREFERRED_URL_SCHEME')

    try:
        info = next(discover_collections(repository, namespace=namespace, name=collection, version=version))
    except StopIteration:
        if upstream and (not no_proxy or namespace not in no_proxy):
            proxy = ProxyUpstream(repository, upstream, cache_read, cache_write, cache_minutes)
            upstream_result = proxy.proxy(request)
            upstream_result['download_url'] = url_for(
                'download.download',
                filename=upstream_result['artifact']['filename'],
                _external=True,
                _scheme=scheme,
                **{C.QUERY_DOWNLOAD_UPSTREAM_URL: upstream_result['download_url']},
            )
            return upstream_result
        else:
            abort(C.HTTP_NOT_FOUND)

    out = {
        'artifact': {
            'filename': info.filename,
            'sha256': info.sha256,
            'size': info.size,
        },
        'collection': {
            'href': url_for(
                ".collection",
                namespace=info.namespace,
                collection=info.name,
                _external=True,
                _scheme=scheme,
            ),
            'name': info.name,
        },
        'namespace': {
            'name': info.namespace,
        },
        'download_url': url_for(
            'download.download',
            filename=info.filename,
            _external=True,
            _scheme=scheme,
        ),
        'hidden': False,
        'href': url_for(
            ".collection",
            namespace=info.namespace,
            collection=info.name,
            _external=True,
            _scheme=scheme,
        ),
        'id': 0,
        'metadata': info.collection_info,
        'version': info.version,
    }
    return out


@v2.route('/collections', methods=['POST'])
@v2.route('/collections/', methods=['POST'])
def publish():
    sha256 = request.form['sha256']
    file = request.files['file']
    skip_configured_auth = current_app.config['PUBLISH_SKIP_CONFIGURED_AUTH']
    property_fallback = current_app.config.get('USE_PROPERTY_FALLBACK', False)
    upload_format = current_app.config.get('UPLOAD_FORMAT')
    _scheme = current_app.config.get('PREFERRED_URL_SCHEME')

    target = authorize(request, current_app.config['ARTIFACTORY_PATH'] / file.filename, skip_configured_auth=skip_configured_auth)

    with _chunk_to_temp(IncomingCollectionStream(file, format=upload_format)) as tmp:
        if tmp.sha256 != sha256:
            abort(Response(f"Hash mismatch: uploaded=='{sha256}', calculated=='{tmp.sha256}'", C.HTTP_INTERNAL_SERVER_ERROR))

        upload_collection_from_hashed_tempfile(target, tmp, property_fallback=property_fallback)

    return jsonify(task=url_for(".import_singleton", _external=True, _scheme=_scheme))



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

import pytest

from galactory import constants as C
from galactory.health import API_RESPONSE

from datetime import datetime

from flask import current_app


def test_health_api(client, trailer):
    response = client.get(trailer('/health'))
    data = response.json

    assert response.status_code == C.HTTP_OK
    assert data == API_RESPONSE
    assert 'v1' in data['available_versions']
    assert 'v1' == data['current_version']


@pytest.mark.parametrize('custom', [None, '', 'test'])
def test_health_basic(client, trailer, custom, app_request_context):
    if custom is None:
        expected = ''
    else:
        current_app.config['HEALTH_CHECK_CUSTOM_TEXT'] = expected = custom

    response = client.get(trailer('/health/v1/basic'))
    data = response.json

    assert response.status_code == C.HTTP_OK
    assert data['type'] == 'basic'
    assert data['custom_text'] == expected
    assert isinstance(data['elapsed_ms'], int)
    assert datetime.fromisoformat(data['response_time']).utcoffset() is None



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from flask import Blueprint, jsonify

from .v1 import bp as v1

API_RESPONSE = {
    'available_versions': {
        'v1': 'v1/',
    },
    'current_version': 'v1',
    'description': 'Galactory Health Checks',
}

bp = Blueprint('health', __name__, url_prefix='/health')
bp.register_blueprint(v1)

@bp.route('')
@bp.route('/')
def health():
    return jsonify(API_RESPONSE)



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from flask import Blueprint

bp = Blueprint('v1', __name__, url_prefix='/v1')
bp.get_send_file_max_age = lambda x: 0

from . import(
    health_checks,
)



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from flask import Blueprint, Response

bp = Blueprint('root', __name__, url_prefix='/')

@bp.route('')
@bp.route('/')
def index():
    return Response('Galactory is running', content_type='text/plain')



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

import pytest
import json
import sys

from pathlib import Path
from unittest import mock
from shutil import copytree
from artifactory import _ArtifactoryAccessor, _FakePathTemplate, ArtifactoryPath

from galactory import create_app

from galactory.utilities import discover_collections as original_discover_collections


@pytest.fixture
def app(request: pytest.FixtureRequest):
    config = getattr(request, 'param', None) or {}
    return create_app(**config)


@pytest.fixture
def app_request_context(app):
    with app.app_context(), app.test_request_context():
        yield


@pytest.fixture
def client(app):
    return app.test_client()


@pytest.fixture
def virtual_fs_repo(fixture_finder, tmp_path: Path):
    repo = tmp_path / 'repo'
    copytree(fixture_finder('artifactory', 'virtual'), repo)
    return repo


@pytest.fixture
def mock_artifactory_accessor(fixture_loader, virtual_fs_repo: Path):
    class MockArtifactoryAccessor(_ArtifactoryAccessor):
        def __init__(self) -> None:
            super().__init__()
            self._galactory_fixture_loader = fixture_loader

        def get_stat_json(self, pathobj, key=None):
            loader = self._galactory_fixture_loader
            repo = pathobj._galactory_mocked_path

            if pathobj.is_dir():
                data = loader.load_json('artifactory/responses/stat_dir')
                # children = repo.listdir(repo)
                data['children'] = [dict(uri='/' + child.name) for child in repo.iterdir()]
                data['uri'] = str(pathobj)
                data['repo'] = virtual_fs_repo.name
                data['path'] = data['uri'].split(f"/{data['repo']}/", 1)[1]
            else:
                data = loader.load_json('artifactory/responses/stat_file')
                data['downloadUri'] = data['uri'] = str(pathobj)
                data['repo'] = virtual_fs_repo.name
                data['path'] = data['uri'].split(f"/{data['repo']}/", 1)[1]

            return data

    return MockArtifactoryAccessor


@pytest.fixture
def mock_artifactory_path(mock_artifactory_accessor, virtual_fs_repo: Path):
    _artifactory_accessor = mock_artifactory_accessor()

    class MockArtifactoryPath(ArtifactoryPath):
        def _new__new__(cls, *args, **kwargs):
            obj = super().__new__(cls, *args, **kwargs)
            rel = str(obj).split('/repo/', 1)[1]
            obj._galactory_mocked_path = virtual_fs_repo / rel
            obj._n = True
            return obj

        def _new_make_child(self, args):
            obj = super()._make_child(args)
            return self._galactory_copy_attrs(obj)

        def _new_make_child_relpath(self, args):
            obj = super()._make_child_relpath(args)
            return self._galactory_copy_attrs(obj)

        if sys.version_info.major == 3 and sys.version_info.minor >= 10:
            _accessor = _artifactory_accessor
            __new__ = _new__new__
            _make_child = _new_make_child
            _make_child_relpath = _new_make_child_relpath
        else:
            # in 3.9 and below Pathlib limits what members can be present in 'Path' class
            __slots__ = ("auth", "verify", "cert", "session", "timeout", "_galactory_mocked_path")

        def _init(self, *args, **kwargs):
            new = super()._init(*args, template=_FakePathTemplate(_artifactory_accessor), **kwargs)
            rel = str(self).split('/repo/', 1)[1]
            self._galactory_mocked_path = virtual_fs_repo / rel
            return new

        def is_dir(self):
            repo = self._galactory_mocked_path
            if repo.is_dir() and not repo.name.endswith('.tar.gz'):
                return True
            else:
                return False
            # return super().is_dir()

        def _galactory_copy_attrs(self, obj):
            obj._galactory_mocked_path = self._galactory_mocked_path
            return obj

        @property
        def _galactory_manifest_path(self):
            return self._galactory_mocked_path / 'MANIFEST.json'

        def _galactory_get_manifest(self, key=None):
            with self._galactory_manifest_path.open() as f:
                manifest = json.load(f)

            if key:
                return manifest.get(key)
            return manifest

        @property
        def properties(self):
            if self.is_dir():
                return {}

            ci = self._galactory_get_manifest('collection_info')

            return {
                'collection_info': [json.dumps(ci)],
                'fqcn': [f"{ci['namespace']}.{ci['name']}"],
                'namespace': [ci['namespace']],
                'name': [ci['name']],
                'version': [ci['version']],
            }

    return MockArtifactoryPath


@pytest.fixture
def repository(mock_artifactory_path):
    return mock_artifactory_path('http://artifactory.example.com/repo/subpath')


@pytest.fixture
def manifest_loader():
    def _load(repo):
        return repo._galactory_get_manifest()

    loader = mock.Mock(wraps=_load)

    with mock.patch('galactory.utilities.load_manifest_from_archive', loader):
        yield loader


@pytest.fixture
def discover_collections(manifest_loader):
    _discover = mock.Mock(wraps=original_discover_collections)
    with mock.patch('galactory.utilities.discover_collections', _discover):
        yield _discover



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

import logging
import warnings

from flask import Flask, request
from werkzeug.middleware.proxy_fix import ProxyFix
from configargparse import ArgParser, ArgumentError, Action
from artifactory import ArtifactoryPath

from .utilities import DateTimeIsoFormatJSONProvider

from .api import create_blueprint as create_api_blueprint
from .download import bp as download
from .health import bp as health
from .root import bp as root


def create_app(**config):
    app = Flask(__name__)
    app.json = DateTimeIsoFormatJSONProvider(app)
    app.config.update(**config)
    app.register_blueprint(health)
    app.register_blueprint(root)
    app.register_blueprint(create_api_blueprint(app))
    app.register_blueprint(download)

    @app.before_request
    def log():
        if app.config.get('LOG_HEADERS'):
            app.logger.debug('Headers: %s', request.headers)

        if app.config.get('LOG_BODY'):
            app.logger.debug('Body: %s', request.get_data())

    return app


# TODO: when py3.8 support is dropped, switch to using argparse.BooleanOptionalAction
class _StrBool(Action):
    FALSES = {'false', '0', 'no'}
    TRUES = {'true', '1', 'yes'}

    def _booler(self, value):
        if isinstance(value, bool):
            return value

        if value.lower() in self.FALSES:
            return False
        if value.lower() in self.TRUES:
            return True

        raise ArgumentError(self, f"Expecting 'true', 'false', 'yes', 'no', '1' or '0', but got '{value}'")

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self._booler(values))


def create_configured_app(run=False, parse_known_only=True, parse_allow_abbrev=False, proxy_fix=False, **extra):
    parser = ArgParser(
        prog='python -m galactory',
        description=(
            'galactory is a partial Ansible Galaxy proxy that uploads and downloads collections, '
            'using an Artifactory generic repository as its backend.'
        ),
        default_config_files=['/etc/galactory.d/*.conf', '~/.galactory/*.conf'],
        allow_abbrev=parse_allow_abbrev,
    )
    parser.add_argument('-c', '--config', required=False, is_config_file=True, env_var='GALACTORY_CONFIG', help='The path to a config file.')
    parser.add_argument('--listen-addr', default='0.0.0.0', type=str, env_var='GALACTORY_LISTEN_ADDR', help='The IP address to listen on.')
    parser.add_argument('--listen-port', default=5555, type=int, env_var='GALACTORY_LISTEN_PORT', help='The TCP port to listen on.')
    parser.add_argument('--server-name', type=str, env_var='GALACTORY_SERVER_NAME', help='The host name and port of the server, as seen from clients. Used for generating links.')
    parser.add_argument('--preferred-url-scheme', type=str, env_var='GALACTORY_PREFERRED_URL_SCHEME', help='Sets the preferred scheme to use when constructing URLs. Defaults to the request scheme, but is unaware of reverse proxies.')
    parser.add_argument('--artifactory-path', type=str, required=True, env_var='GALACTORY_ARTIFACTORY_PATH', help='The URL of the path in Artifactory where collections are stored.')
    parser.add_argument('--artifactory-api-key', type=str, env_var='GALACTORY_ARTIFACTORY_API_KEY', help='If set, is the API key used to access Artifactory. If set with artifactory-access-token, this value will not be used.')
    parser.add_argument('--artifactory-access-token', type=str, env_var='GALACTORY_ARTIFACTORY_ACCESS_TOKEN', help='If set, is the Access Token used to access Artifactory. If set with artifactory-api-key, this value will be used and the API key will be ignored.')
    parser.add_argument('--use-galaxy-key', action='store_true', env_var='GALACTORY_USE_GALAXY_KEY', help='If set, uses the Galaxy token sent in the request as the Artifactory auth. DEPRECATED: This option will be removed in v0.12.0. Please use --use-galaxy-auth going forward.')
    parser.add_argument('--use-galaxy-auth', action='store_true', env_var='GALACTORY_USE_GALAXY_AUTH', help='If set, uses the Galaxy token sent in the request as the Artifactory auth.')
    parser.add_argument('--galaxy-auth-type', type=str, env_var='GALACTORY_GALAXY_AUTH_TYPE', choices=['api_key', 'access_token'], help='Auth received via a Galaxy request should be interpreted as this type of auth.')
    parser.add_argument('--prefer-configured-key', action='store_true', env_var='GALACTORY_PREFER_CONFIGURED_KEY', help='If set, prefer the confgured Artifactory auth over the Galaxy token. DEPRECATED: This option will be removed in v0.12.0. Please use --prefer-configured-auth going forward.')
    parser.add_argument('--prefer-configured-auth', action='store_true', env_var='GALACTORY_PREFER_CONFIGURED_AUTH', help='If set, prefer the confgured Artifactory auth over the Galaxy token.')
    parser.add_argument('--publish-skip-configured-key', action='store_true', env_var='GALACTORY_PUBLISH_SKIP_CONFIGURED_KEY', help='If set, publish endpoint will not use configured auth, only auth included in a Galaxy request. DEPRECATED: This option will be removed in v0.12.0. Please use --publish-skip-configured-auth going forward.')
    parser.add_argument('--publish-skip-configured-auth', action='store_true', env_var='GALACTORY_PUBLISH_SKIP_CONFIGURED_AUTH', help='If set, publish endpoint will not use configured auth, only auth included in a Galaxy request.')
    parser.add_argument('--log-file', type=str, env_var='GALACTORY_LOG_FILE', help='If set, logging will go to this file instead of the console.')
    parser.add_argument(
        '--log-level',
        type=str.upper,
        env_var='GALACTORY_LOG_LEVEL',
        default='INFO',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], help='The desired logging level.',
    )
    parser.add_argument('--log-headers', action='store_true', env_var='GALACTORY_LOG_HEADERS', help='Log the headers of every request (DEBUG level only).')
    parser.add_argument('--log-body', action='store_true', env_var='GALACTORY_LOG_BODY', help='Log the body of every request (DEBUG level only).')
    parser.add_argument('--proxy-upstream', type=lambda x: str(x).rstrip('/') + '/', env_var='GALACTORY_PROXY_UPSTREAM', help='If set, then find, pull and cache results from the specified galaxy server in addition to local.')
    parser.add_argument('-npns', '--no-proxy-namespace', action='append', default=[], env_var='GALACTORY_NO_PROXY_NAMESPACE', help='Requests for this namespace should never be proxied. Can be specified multiple times.')
    parser.add_argument('--cache-minutes', default=60, type=int, env_var='GALACTORY_CACHE_MINUTES', help='The time period that a cache entry should be considered valid.')
    parser.add_argument('--cache-read', action=_StrBool, default=True, env_var='GALACTORY_CACHE_READ', help='Look for upsteam caches and use their values.')
    parser.add_argument('--cache-write', action=_StrBool, default=True, env_var='GALACTORY_CACHE_WRITE', help='Populate the upstream cache in Artifactory. Should be false when no auth is provided or the auth has no permission to write.')
    parser.add_argument('--use-property-fallback', action='store_true', env_var='GALACTORY_USE_PROPERTY_FALLBACK', help='Set properties of an uploaded collection in a separate request after publshinng. Requires a Pro license of Artifactory. This feature is a workaround for an Artifactory proxy configuration error and may be removed in a future version.')
    parser.add_argument('--health-check-custom-text', type=str, default='', env_var='GALACTORY_HEALTH_CHECK_CUSTOM_TEXT', help='Sets custom_text field for health check endpoint responses.')
    parser.add_argument('--api-version', action='append', choices=['v2', 'v3'], env_var='GALACTORY_API_VERSION', help='The API versions to serve. Can be set to limit functionality to specific versions only. Defaults to all supported versions.')
    parser.add_argument('--upload-format', type=str, env_var='GALACTORY_UPLOAD_FORMAT', choices=['base64', 'raw', 'auto'], default='auto', help='Galaxy accepts the uploaded collection tarball as either raw bytes or base64 encoded. Ansible 2.9 uploads raw bytes, later versions upload base64. By default galactory will try to auto-detect. Use this option to turn off auto-detection and force a specific format.')

    if parse_known_only:
        args, _ = parser.parse_known_args()
    else:
        args = parser.parse_args()

    logging.basicConfig(filename=args.log_file, level=args.log_level)

    # TODO: v0.12.0 - remove conditional old name
    if args.use_galaxy_key and not args.use_galaxy_auth:
        use_galaxy_auth = True
        warnings.warn(
            message=(
                "USE_GALAXY_KEY has been replaced by USE_GALAXY_AUTH and the old name will be removed in v0.12.0."
                " To suppress this warning, set USE_GALAXY_AUTH."
            ), category=DeprecationWarning, stacklevel=2
        )
    else:
        use_galaxy_auth = args.use_galaxy_auth

    # TODO: v0.12.0 - remove conditional & warning, set default on argument
    if args.galaxy_auth_type is None and use_galaxy_auth:
        galaxy_auth_type = 'api_key'
        warnings.warn(
            message=(
                "USE_GALAXY_AUTH is True but GALAXY_AUTH_TYPE is not set."
                " The default value used will be 'api_key' for backward compatibility, but will change to 'access_token' in v0.12.0."
                " To suppress this warning, set an explicit value."
            ), category=FutureWarning, stacklevel=2
        )
    else:
        galaxy_auth_type = args.galaxy_auth_type

    # TODO: v0.12.0 - remove conditional old name
    if args.prefer_configured_key and not args.prefer_configured_auth:
        prefer_configured_auth = True
        warnings.warn(
            message=(
                "PREFER_CONFIGURED_KEY has been replaced by PREFER_CONFIGURED_AUTH and the old name will be removed in v0.12.0."
                " To suppress this warning, set PREFER_CONFIGURED_AUTH."
            ), category=DeprecationWarning, stacklevel=2
        )
    else:
        prefer_configured_auth = args.prefer_configured_auth

    # TODO: v0.12.0 - remove conditional old name
    if args.publish_skip_configured_key and not args.publish_skip_configured_auth:
        publish_skip_configured_auth = True
        warnings.warn(
            message=(
                "PUBLISH_SKIP_CONFIGURED_KEY has been replaced by PUBLISH_SKIP_CONFIGURED_AUTH and the old name will be removed in v0.12.0."
                " To suppress this warning, set PUBLISH_SKIP_CONFIGURED_AUTH."
            ), category=DeprecationWarning, stacklevel=2
        )
    else:
        publish_skip_configured_auth = args.publish_skip_configured_auth

    app = create_app(
        ARTIFACTORY_PATH=ArtifactoryPath(args.artifactory_path),
        LOG_HEADERS=args.log_headers,
        LOG_BODY=args.log_body,
        PROXY_UPSTREAM=args.proxy_upstream,
        NO_PROXY_NAMESPACES=args.no_proxy_namespace,
        ARTIFACTORY_API_KEY=args.artifactory_api_key,
        ARTIFACTORY_ACCESS_TOKEN=args.artifactory_access_token,
        USE_GALAXY_AUTH=use_galaxy_auth,
        GALAXY_AUTH_TYPE=galaxy_auth_type,
        PREFER_CONFIGURED_AUTH=prefer_configured_auth,
        PUBLISH_SKIP_CONFIGURED_AUTH=publish_skip_configured_auth,
        SERVER_NAME=args.server_name,
        PREFERRED_URL_SCHEME=args.preferred_url_scheme,
        CACHE_MINUTES=args.cache_minutes,
        CACHE_READ=args.cache_read,
        CACHE_WRITE=args.cache_write,
        USE_PROPERTY_FALLBACK=args.use_property_fallback,
        HEALTH_CHECK_CUSTOM_TEXT=args.health_check_custom_text,
        API_VERSION=args.api_version,
        UPLOAD_FORMAT=args.upload_format,
    )

    if proxy_fix:
        proxy_args = {k: v for k, v in extra.items() if k.startswith('x_')}
        app = ProxyFix(app, **proxy_args)

    if run:
        app.run(args.listen_addr, args.listen_port, threaded=True)

    return app



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

import pytest
from unittest import mock

from datetime import datetime, timedelta
from types import GeneratorType
import semver

from galactory.utilities import discover_collections
from galactory.models import CollectionData


def test_discover_collections_skip_dirs(repository):
    with mock.patch('artifactory.ArtifactoryFileStat', mock.Mock(return_value=mock.Mock(is_dir=True, children=[]))):
        collections = list(discover_collections(repository))
        assert collections == []


@pytest.mark.parametrize('props', [{}, {'version': []}])
def test_discover_collections_skip_missing_version(repository, props):
    with mock.patch.object(repository.__class__, 'properties', property(mock.Mock(return_value=props))):
        collections = list(discover_collections(repository))
        assert collections == []


@pytest.mark.parametrize('namespace', [None, 'community', 'briantist', 'fake'])
@pytest.mark.parametrize('collection', [None, 'whatever', 'hashi_vault', 'fake', 'devel'])
@pytest.mark.parametrize('version', [None, '2.5.0', '3.0.0', '0.1.0', '0.2.0', '0.0.0'])
def test_discover_collections_any(repository, manifest_loader, namespace, collection, version, app_request_context):
    gen = discover_collections(repository, namespace, collection, version)

    assert isinstance(gen, GeneratorType)

    collections = list(gen)
    contents = list(repository)

    assert len(collections) <= len(contents)
    if any([
        namespace == 'fake',
        collection == 'fake',
        version == '0.0.0',
    ]):
        assert len(collections) == 0

    # TODO: we're phasing out the manifest loader, remove later
    # assert len(contents) == manifest_loader.call_count
    # expected_calls = [mock.call(x) for x in contents]
    # manifest_loader.assert_has_calls(expected_calls, any_order=True)

    for c in collections:
        assert isinstance(c, CollectionData)
        assert c.created_datetime.utcoffset() == timedelta(0)
        assert c.modified_datetime.utcoffset() == timedelta(0)
        assert datetime.fromisoformat(c.created) == c.created_datetime
        assert datetime.fromisoformat(c.modified) == c.modified_datetime
        assert c.fqcn == f"{c.namespace}.{c.name}"
        assert isinstance(c.semver, semver.VersionInfo)
        assert semver.VersionInfo.parse(c.version) == c.semver

        assert namespace is None or c.namespace == namespace
        assert collection is None or c.name == collection
        assert version is None or c.version == version



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from datetime import datetime

from . import bp

@bp.route('/collection-imports/0')
@bp.route('/collection-imports/0/')
def import_singleton():
    out = {
        'state': 'completed',
        'finished_at': datetime.utcnow(),
    }
    return out



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from galactory import constants as C


# can't use trailer fixture here
# for some reason with internal client,
# .get('') gives a 308 redirect, even
# though that doesn't happen with a real
# webserver and request.
def test_root(client):
    response = client.get('/')
    data = response.text

    assert response.status_code == C.HTTP_OK
    assert data == 'Galactory is running'



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

import pytest
from galactory import constants as C

from datetime import datetime

@pytest.mark.parametrize('app', [
    dict(API_VERSION=['v2']),
    dict(API_VERSION=['v3']),
    dict(API_VERSION=['v3', 'v2']),
    None,
], indirect=True)
def test_api(app, client, trailer):
    response = client.get(trailer('/api'))
    data = response.json

    possible_versions = ('v2', 'v3')
    api_config = app.config.get('API_VERSION')
    active_versions = api_config or possible_versions

    assert response.status_code == C.HTTP_OK
    assert 'available_versions' in data

    if 'v2' in active_versions:
        assert data['current_version'] == 'v2'
    else:
        assert 'current_version' not in data

    for v in active_versions:
        assert v in possible_versions
        assert v in data['available_versions']
        assert data['available_versions'][v] == f"{v}/"


@pytest.mark.parametrize(['app', 'path'], [
    (dict(API_VERSION=['v2']), 'collection-imports/0'),
    (dict(API_VERSION=['v3']), 'imports/collections/0'),
], indirect=['app'])
def test_import_collection_status(app, path, client, trailer):
    active_version = app.config['API_VERSION'][0]
    response = client.get(trailer(f"/api/{active_version}/{path}"))
    data = response.json

    assert response.status_code == C.HTTP_OK
    assert data['state'] == 'completed'
    assert datetime.fromisoformat(data['finished_at']).utcoffset() is None



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from datetime import datetime
from time import perf_counter
from flask import jsonify, g as ctx, current_app, url_for, request

from . import bp


@bp.before_request
def start_timer():
    ctx.start_time = perf_counter()


@bp.after_request
def commonize(response):
    elapsed_ms = int((perf_counter() - ctx.start_time) * 1000)

    scheme = current_app.config.get('PREFERRED_URL_SCHEME')
    custom_txt = current_app.config.get('HEALTH_CHECK_CUSTOM_TEXT', '')

    data = response.get_json()

    data['elapsed_ms'] = elapsed_ms
    data['response_time'] = datetime.utcnow()
    data['custom_text'] = custom_txt
    data['href'] = url_for(request.endpoint, _external=True, _scheme=scheme, **request.view_args)

    response.data = current_app.json.dumps(data)

    response.headers['Cache-Control'] = "no-cache, no-store, must-revalidate"
    response.headers['Pragma'] = 'no-cache'
    response.headers['Expires'] = '0'

    return response


@bp.route('/basic')
@bp.route('/basic/')
def basic():
    out = {
        'type': 'basic'
    }
    return jsonify(out)


#TODO: add more healthcheck types?



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

from flask import Blueprint, Flask


def create_blueprint(app: Flask):
    api_version = app.config.get('API_VERSION')
    response = {
        'available_versions': {},
        'description': 'GALAXY REST API',
    }

    bp = Blueprint('api', __name__, url_prefix='/api')

    if api_version is None or 'v3' in api_version:
        app.logger.info("Registering Galaxy API v3")
        response['available_versions']['v3'] = 'v3/'
        from .v3 import bp as v3
        bp.register_blueprint(v3)

    if api_version is None or 'v2' in api_version:
        app.logger.info("Registering Galaxy API v2")
        response['available_versions']['v2'] = 'v2/'
        response['current_version'] = 'v2' # This field doesn't exist in the v3 output.
        from .v2 import bp as v2
        bp.register_blueprint(v2)

    @bp.route('')
    @bp.route('/', endpoint='api')
    def api():
        return response

    return bp



# -*- coding: utf-8 -*-
# (c) 2023 Brian Scholer (@briantist)

import pytest
import tarfile

from pathlib import Path
from base64io import Base64IO
from galactory.utilities import IncomingCollectionStream


@pytest.fixture
def collection_tarball(virtual_fs_repo: Path, tmp_path: Path):
    collection = next(virtual_fs_repo.glob("**/*.tar.gz"))
    gz_path = tmp_path / collection.name
    with tarfile.open(gz_path, mode='w:gz') as tar:
        tar.add(collection)
    return gz_path


@pytest.fixture
def base64_tarball(collection_tarball: Path, tmp_path: Path):
    b64_path = tmp_path / f"{collection_tarball.name}.b64"
    with open(collection_tarball, mode='rb') as raw, open(b64_path, mode='wb') as w, Base64IO(w) as f:
        f.write(raw.read())
    return b64_path


class TestIncomingCollectionStream:
    @pytest.mark.parametrize('format', [None, 'auto', 'undefined', 'raw'])
    def test_raw(self, collection_tarball: Path, format: str):
        with open(collection_tarball, mode='rb') as f:
            assert IncomingCollectionStream.detected_stream(f) is f
            assert IncomingCollectionStream(f, format=format)

    @pytest.mark.parametrize('format', [None, 'auto', 'undefined', 'base64'])
    def test_base64(self, base64_tarball: Path, format: str):
        with open(base64_tarball, mode='rb') as f:
            assert isinstance(IncomingCollectionStream.detected_stream(f), Base64IO)
            assert isinstance(IncomingCollectionStream(f, format=format), Base64IO)



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

import json
import math
import hashlib
import gzip

import typing as t

from datetime import datetime
from tempfile import SpooledTemporaryFile
from urllib3 import Retry
from requests.adapters import HTTPAdapter
from requests import Session
from base64io import Base64IO

from flask import current_app, abort, Response, Request
from flask.json.provider import DefaultJSONProvider
from artifactory import ArtifactoryPath, ArtifactoryException
from dohq_artifactory.auth import XJFrogArtApiAuth, XJFrogArtBearerAuth

from . import constants as C
from .models import CollectionData
from .iter_tar import iter_tar


class DateTimeIsoFormatJSONProvider(DefaultJSONProvider):
    @staticmethod
    def default(o: t.Any) -> t.Any:
        if isinstance(o, datetime):
            return o.isoformat()

        return super().default(o)


def _session_with_retries(retry=None, auth=None) -> Session:
    if retry is None:
        retry = Retry(connect=5, read=3, redirect=2, status=6, other=3, backoff_factor=0.1, raise_on_status=False)

    adapter = HTTPAdapter(max_retries=retry)
    session = Session()
    session.auth = auth
    session.mount('http://', adapter)
    session.mount('https://', adapter)

    return session


def authorize(request: Request, artifactory_path: ArtifactoryPath, retry=None, skip_configured_auth: bool = False) -> ArtifactoryPath:
    auth = None
    if not skip_configured_auth:
        accesstoken = current_app.config['ARTIFACTORY_ACCESS_TOKEN']
        apikey = current_app.config['ARTIFACTORY_API_KEY']
        if accesstoken is not None:
            auth = XJFrogArtBearerAuth(accesstoken)
        elif apikey is not None:
            auth = XJFrogArtApiAuth(apikey)

    if current_app.config['USE_GALAXY_AUTH'] and (not current_app.config['PREFER_CONFIGURED_AUTH'] or auth is None):
        galaxy_auth_type = current_app.config['GALAXY_AUTH_TYPE']
        authorization = request.headers.get('Authorization')
        if authorization:
            token = authorization.split(' ')[1]
            if galaxy_auth_type == 'access_token':
                auth = XJFrogArtBearerAuth(token)
            elif galaxy_auth_type == 'api_key':
                auth = XJFrogArtApiAuth(token)
            else:
                raise ValueError(f"Unknown galaxy auth type '{galaxy_auth_type}'.")

    session = _session_with_retries(retry=retry, auth=auth)
    return ArtifactoryPath(artifactory_path, session=session)


def load_manifest_from_archive(handle, seek_to_zero_after=True):
    with gzip.GzipFile(fileobj=handle, mode='rb') as gz:
        for fname, data in iter_tar(gz):
            if fname.lower() in ('manifest.json', './manifest.json'):
                data = json.loads(data)
                if seek_to_zero_after:
                    handle.seek(0)
                return data


def discover_collections(
    repo: ArtifactoryPath,
    namespace: str = None,
    name: str = None,
    version: str = None,
    fast_detection: bool = True,
):
    for p in repo:
        if fast_detection:
            # we're going to use the naming convention to eliminate candidates early,
            # to avoid excessive additional requests for properties and stat that slow
            # down the listing immensely as the number of collections grows.
            try:
                f_namespace, f_name, f_version = p.name.replace('.tar.gz', '').split('-', maxsplit=2)
            except ValueError:
                pass
            else:
                if not all(
                    (
                        not namespace or f_namespace == namespace,
                        not name or f_name == name,
                        not version or f_version == version
                    )
                ):
                    continue

        info = p.stat()
        if info.is_dir:
            continue

        props = p.properties
        if not props.get('collection_info'):
            continue

        coldata = CollectionData.from_artifactory_path(path=p, properties=props, stat=info)

        if all(
            (
                not namespace or coldata.namespace == namespace,
                not name or coldata.name == name,
                not version or coldata.version == version
            )
        ):
            yield coldata


def lcm(a, b, *more):
    z = lcm(b, *more) if more else b
    return abs(a * z) // math.gcd(a, z)


class IncomingCollectionStream:
    def __new__(cls, stream: t.IO, *, format: str = None):
        if format == 'raw':
            return stream
        if format == 'base64':
            return Base64IO(stream)
        return cls.detected_stream(stream)

    @staticmethod
    def detected_stream(stream: t.IO):
        with gzip.GzipFile(fileobj=stream, mode='rb') as gz:
            try:
                gz.read(1)
            except gzip.BadGzipFile:
                return Base64IO(stream)
            else:
                return stream
            finally:
                stream.seek(0)


class HashedTempFile():
    def __init__(self, handle, md5, sha1, sha256, close=True) -> None:
        self.handle = handle
        self.md5 = md5
        self.sha1 = sha1
        self.sha256 = sha256
        self._close = close

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self._close:
            self.handle.close()


def _chunk_to_temp(fsrc, iterator=None, spool_size=5*1024*1024, seek_to_zero=True, chunk_multiplier=64, close=True) -> HashedTempFile:
    md5sum = hashlib.md5()
    sha1sum = hashlib.sha1()
    sha256sum = hashlib.sha256()
    common_block_size = lcm(md5sum.block_size, sha1sum.block_size, sha256sum.block_size)
    chunk_size = chunk_multiplier * common_block_size

    it = iter(lambda: fsrc.read(chunk_size), b'') if iterator is None else iterator(chunk_size)

    tmp = SpooledTemporaryFile(max_size=spool_size)

    for chunk in it:
        md5sum.update(chunk)
        sha1sum.update(chunk)
        sha256sum.update(chunk)
        tmp.write(chunk)

    if seek_to_zero:
        tmp.seek(0)

    return HashedTempFile(tmp, md5sum.hexdigest(), sha1sum.hexdigest(), sha256sum.hexdigest(), close=close)


def upload_collection_from_hashed_tempfile(artifact: ArtifactoryPath, tmpfile: HashedTempFile, property_fallback: bool = False) -> t.Dict[str, t.Any]:
    try:
        manifest = load_manifest_from_archive(tmpfile.handle)
    except Exception:
        abort(Response("Error loading manifest from collection archive.", C.HTTP_INTERNAL_SERVER_ERROR))
    else:
        ci = manifest['collection_info']
        props = {
            'collection_info': json.dumps(ci),
            'namespace': ci['namespace'],
            'name': ci['name'],
            'version': ci['version'],
            'fqcn': f"{ci['namespace']}.{ci['name']}"
        }

    params = props
    if property_fallback:
        params = {}

    try:
        artifact.deploy(
            tmpfile.handle,
            md5=tmpfile.md5,
            sha1=tmpfile.sha1,
            sha256=tmpfile.sha256,
            parameters=params,
            quote_parameters=True
        )
    except ArtifactoryException as exc:
        cause = exc.__cause__
        current_app.logger.debug(cause)
        abort(Response(cause.response.text, cause.response.status_code))
    else:
        if property_fallback:
            artifact.properties = props

    return props



# -*- coding: utf-8 -*-
# (c) 2023 Brian Scholer (@briantist)

import pytest

import re
from semver import VersionInfo
from pytest_mock import MockFixture

from galactory.models import CollectionData, CollectionGroup


@pytest.mark.parametrize('name', ['collection_name', 'name2'])
@pytest.mark.parametrize('namespace', ['a_namespace', 'ns1'])
def test_collectiongroup_init(namespace, name):
    colgroup = CollectionGroup(namespace=namespace, name=name)

    assert isinstance(colgroup, CollectionGroup)
    assert colgroup.namespace == namespace
    assert colgroup.name == name
    assert colgroup.fqcn == f"{namespace}.{name}"
    assert colgroup.latest is None
    assert len(colgroup) == len(colgroup.versions) == 0


@pytest.mark.parametrize('name', ['collection_name'])
@pytest.mark.parametrize('namespace', ['a_namespace'])
@pytest.mark.parametrize('version', [
    'abc.0',
    '0.hello',
    '99999.0.1111.6.4.3',
    None,
    4.4,
    True,
    [],
    {},
])
@pytest.mark.parametrize('value', [None, "thing", 2, 3.14159, False, [], {}])
def test_collectiongroup_bad_entries(mocker: MockFixture, namespace, name, collection_data, version, value):
    colgroup = CollectionGroup(namespace=namespace, name=name)

    # bad value type
    with pytest.raises(TypeError, match=r'^Values must be of type CollectionData\.$'):
        colgroup[version] = value

    # bad collection name
    with pytest.raises(ValueError, match=rf"^Attempted to add collection '{collection_data.namespace}.{collection_data.name}' to group for '{namespace}\.{name}'\.$"):
        colgroup[version] = collection_data

    # bad key (version)
    collection_data.namespace = namespace
    collection_data.name = name
    regver = re.escape(str(version))
    spy = mocker.spy(colgroup, '_get_key')
    with pytest.raises((TypeError, ValueError), match=rf"^(?:Only valid semantic versions can be used as keys\.|{regver} is not valid SemVer string)$"):
        colgroup[version] = collection_data

    spy.assert_called_once_with(version)


@pytest.mark.parametrize('name', ['collection_name'])
@pytest.mark.parametrize('namespace', ['a_namespace'])
def test_collectiongroup_add(mocker: MockFixture, namespace, name, collection_data):
    spy = mocker.spy(CollectionGroup, '__setitem__')
    colgroup = CollectionGroup(namespace=namespace, name=name)

    collection_data.namespace = namespace
    collection_data.name = name


    colgroup.add(collection_data)
    spy.assert_called_once_with(colgroup, collection_data.semver, collection_data)
    assert len(colgroup) == 1
    assert colgroup.latest is collection_data


def test_collectiongroup_from_collection(mocker: MockFixture, collection_data: CollectionData):
    spy_init = mocker.spy(CollectionGroup, '__init__')
    spy_add = mocker.spy(CollectionGroup, 'add')
    colgroup = CollectionGroup.from_collection(collection_data)

    spy_init.assert_called_once_with(colgroup, namespace=collection_data.namespace, name=collection_data.name)
    spy_add.assert_called_once_with(colgroup, collection_data)
    assert len(colgroup) == 1
    assert colgroup.latest is collection_data


def test_collectiongroup_dunders(mocker: MockFixture, collection_data: CollectionData):
    colgroup = CollectionGroup.from_collection(collection_data)
    spy_get_key = mocker.spy(colgroup, '_get_key')

    assert colgroup[collection_data.semver] is collection_data
    spy_get_key.assert_called_once_with(collection_data.semver)
    spy_get_key.reset_mock()

    assert collection_data.version in colgroup
    spy_get_key.assert_called_once_with(collection_data.version, raises=False)
    spy_get_key.reset_mock()


def test_collectiongroup_delete(mocker: MockFixture, collection_data_factory):
    spy_del = mocker.spy(CollectionGroup, '__delitem__')

    col1 = collection_data_factory(version='1.2.3', sha256='A')
    col2 = collection_data_factory(version='2.3.4', sha256='B')

    assert col1 is not col2
    assert col1 != col2
    assert col1 < col2
    assert col2 > col1

    colgroup = CollectionGroup.from_collection(col1)
    assert colgroup.latest is col1

    colgroup.add(col2)
    assert len(colgroup) == 2
    assert col1 in colgroup.values()
    assert col2 in colgroup.values()
    assert colgroup.latest is col2

    del(colgroup[col2.version])
    spy_del.assert_called_once_with(colgroup, col2.version)
    spy_del.reset_mock()
    assert colgroup.latest is col1

    del(colgroup[col1.semver])
    spy_del.assert_called_once_with(colgroup, col1.semver)
    assert colgroup.latest is None


@pytest.mark.parametrize('version', ['1.2.3-dev0', '2.3.4', '0.0.0-alpha'])
@pytest.mark.parametrize('as_vi', [True, False])
def test_collectiongroup_get_key_good_versions(version, as_vi):
    v = VersionInfo.parse(version)
    if as_vi:
        k = CollectionGroup._get_key(v)
        assert k is v
    else:
        k = CollectionGroup._get_key(version)
        assert k == v


@pytest.mark.parametrize('version', ['1.2.3dev0', 'now2.3.4', '0.0.0.0.0.0.0'])
def test_collectiongroup_get_key_bad_version_strings(version):
    with pytest.raises(ValueError, match=rf"^{version} is not valid SemVer string$"):
        CollectionGroup._get_key(version)


@pytest.mark.parametrize('version', [list(), tuple(), dict(), True, 2, 3.14159])
@pytest.mark.parametrize('raises', [True, False])
def test_collectiongroup_get_key_bad_version_types(version, raises):
    if raises:
        with pytest.raises(TypeError, match=r"Only valid semantic versions can be used as keys.$"):
            CollectionGroup._get_key(version, raises=raises)
    else:
        k = CollectionGroup._get_key(version, raises=raises)
        assert k is version



# -*- coding: utf-8 -*-
# (c) 2022 Brian Scholer (@briantist)

import requests
import json

from contextlib import contextmanager
from io import StringIO
from datetime import datetime, timedelta
from artifactory import ArtifactoryException

from flask import current_app, abort, Response, url_for

from . import constants as C
from .utilities import _session_with_retries, DateTimeIsoFormatJSONProvider

class _CacheEntry:
    _raw = {}

    @classmethod
    def from_file(cls, f, **kwargs):
        def _time_decoder(pairs):
            d = {}
            for k, v in pairs:
                try:
                    nv = datetime.fromisoformat(v)
                except (ValueError, TypeError):
                    d[k] = v
                else:
                    d[k] = nv

            return d

        loaded = json.load(f, object_pairs_hook=_time_decoder)
        return cls(data=loaded['data'], metadata=loaded['metadata'], **kwargs)

    def __init__(self, expiry_delta, data=None, metadata=None, calculate_expiry_on_read=True) -> None:
        raw = {'metadata': {}, 'data': {}}
        self._expiry_delta = expiry_delta
        self._calc_on_read = calculate_expiry_on_read
        if data is not None:
            raw['data'] = data.copy()

        if metadata is not None:
            raw['metadata'] = metadata.copy()

        self._raw = raw

    @property
    def empty(self) -> bool:
        return not (self._raw.get('data') and self._raw.get('metadata'))

    @property
    def created(self) -> datetime:
        if self.empty:
            return None

        return self.metadata.get('created')

    @property
    def expires(self) -> datetime:
        if self.empty:
            return None

        if self._calc_on_read:
            if self.created is None:
                return None

            return self.created + self._expiry_delta
        else:
            return self.metadata.get('expires')

    @property
    def expired(self) -> bool:
        expires = self.expires
        return expires is not None and datetime.utcnow() >= expires

    @property
    def metadata(self) -> dict:
        return self._raw['metadata']

    @property
    def data(self) -> dict:
        return self._raw['data']

    @data.setter
    def data(self, value) -> None:
        self._raw['data'] = value
        self.metadata['dirty'] = True

    @property
    def dirty(self) -> bool:
        return self.metadata.get('dirty', False)

    def update(self, force=False):
        if not (force or self.dirty):
            return

        now = datetime.utcnow()
        self.metadata['created'] = now
        self.metadata['expires'] = now + self._expiry_delta
        self.metadata['dirty'] = False

    def _to_serializable_dict(self):
        o = {
            'data': self.data,
            'metadata': self.metadata.copy(),
        }
        o['metadata'].pop('dirty')
        return o


class ProxyUpstream:
    _cache_path = '_cache'

    def __init__(self, repository, upstream_url, read_cache, write_cache, cache_expiry_minutes) -> None:
        self._repository = repository
        self._upstream = upstream_url
        self._read_cache = read_cache
        self._write_cache = write_cache
        self._cache_expiry_delta = timedelta(minutes=cache_expiry_minutes)


    def _get_cache(self, request, expiry_delta=None, **kwargs) -> _CacheEntry:
        path = self._repository / self._cache_path / request.path / 'data.json'

        if expiry_delta is None:
            expiry_delta = self._cache_expiry_delta

        if not self._read_cache:
            return _CacheEntry(expiry_delta=expiry_delta, **kwargs)

        try:
            with path.open() as f:
                return _CacheEntry.from_file(f, expiry_delta=expiry_delta, **kwargs)
        except ArtifactoryException:
            return _CacheEntry(expiry_delta=expiry_delta, **kwargs)

    def _set_cache(self, request, cache) -> None:
        if not self._write_cache:
            return

        path = self._repository / self._cache_path / request.path / 'data.json'
        with StringIO() as buffer:
            cache.update()
            json.dump(cache._to_serializable_dict(), buffer, default=DateTimeIsoFormatJSONProvider.default)
            buffer.seek(0)
            path.deploy(buffer, quote_parameters=True)

    @contextmanager
    def proxy_download(self, request):
        no_rewrite = C.QUERY_DOWNLOAD_UPSTREAM_URL in request.args
        if no_rewrite:
            upstream_url = request.args[C.QUERY_DOWNLOAD_UPSTREAM_URL]
        else:
            upstream_url = self._upstream

        req = self._rewrite_to_upstream(request, upstream_url, no_rewrite=no_rewrite, no_paginate=True)
        with _session_with_retries() as s:
            try:
                # Merge environment settings into session
                settings = s.merge_environment_settings(req.url, proxies={}, stream=True, verify=None, cert=None)
                resp = s.send(req, **settings)
                resp.raise_for_status()
            except:
                abort(Response(resp.text, resp.status_code))
            else:
                try:
                    yield resp
                finally:
                    resp.close()

    def proxy(self, request):
        cache = self._get_cache(request)
        scheme = current_app.config.get('PREFERRED_URL_SCHEME')

        if cache.empty or cache.expired:
            req = self._rewrite_to_upstream(request, self._upstream)
            with _session_with_retries() as s:
                try:
                    # Merge environment settings into session
                    settings = s.merge_environment_settings(req.url, proxies={}, stream=None, verify=None, cert=None)
                    resp = s.send(req, **settings)
                    resp.raise_for_status()
                except requests.exceptions.HTTPError:
                    infoer = current_app.logger.debug if resp.status_code == C.HTTP_NOT_FOUND else current_app.logger.warning
                    infoer("Upstream results not available, got HTTP %i: %s", resp.status_code, resp.text)
                    if cache.expired:
                        current_app.logger.info(f"Cache hit (expired, upstream error): {request.url}")
                        return cache.data
                    # else:
                        # abort(Response(resp.text, resp.status_code))

                else:
                    if self._read_cache:
                        current_app.logger.info(f"Cache miss: {request.url}")

                    data = resp.json()
                    cache.data = data

                    if self._write_cache:
                        self._set_cache(request, cache)
        else:
            current_app.logger.info(f"Cache hit: {request.url}")

        return self._rewrite_upstream_response(cache.data, url_for('root.index', _external=True, _scheme=scheme))

    def _rewrite_upstream_response(self, response_data, url_root) -> dict:
        # Remove these keys from the response.
        # If the value is not None, only remove when the type matches.
        _REMOVE_FIELDS = {
            # TODO: should id refer to some Artifactory ID?
            'id': (int, str),
            # TODO: use artifactory download count, or combine with upstream?
            'download_count': int,
        }

        # Leave these fields in the response without alteration or
        # further processing. Same rules as _REMOVE_FIELDS.
        _SKIP_FIELDS = {
            # This field is an absolute URL, and will not necessarily map to
            # a known API path, even if it does in public Galaxy.
            # We will leave it unadultered and let the API paths that expect
            # this field to overwrite it or not.
            'download_url': None,
        }

        ret = {}
        for k, v in response_data.items():
            if k in _SKIP_FIELDS and (_SKIP_FIELDS[k] is None or isinstance(v, _SKIP_FIELDS[k])):
                ret[k] = v
                continue
            if k in _REMOVE_FIELDS and (_REMOVE_FIELDS[k] is None or isinstance(v, _REMOVE_FIELDS[k])):
                continue

            if isinstance(v, dict):
                ret[k] = self._rewrite_upstream_response(v, url_root)
            elif isinstance(v, list):
                ret[k] = [self._rewrite_upstream_response(d, url_root) if isinstance(d, dict) else d for d in v]
            elif isinstance(v, str) and v.startswith(self._upstream):
                if 'api/v1' in v:
                    continue
                ret[k] = v.replace(self._upstream, url_root)
            else:
                ret[k] = v

        return ret

    def _rewrite_to_upstream(self, request, upstream_url, *, prepared=True, no_rewrite=False, no_paginate=False):
        this_url = request.base_url
        this_root = request.url_root
        if no_rewrite:
            params = {}
            rewritten = upstream_url
            current_app.logger.info(f"Not rewriting '{this_url}'; using '{rewritten}'")
        else:
            rewritten = this_url.replace(this_root, upstream_url)
            current_app.logger.info(f"Rewriting '{this_url}' to '{rewritten}'")
            params = request.args.copy()
            if not no_paginate:
                if 'v2' in this_url:
                    params['page_size'] = 100
                else:
                    params['limit'] = 100

        headers = {k: v for k, v in request.headers.items() if k not in ['Authorization', 'Host']}
        headers['Accept'] = 'application/json, */*'

        req = requests.Request(method=request.method, url=rewritten, headers=headers, data=request.data, params=params)

        if prepared:
            prepared = req.prepare()
            current_app.logger.info(prepared.url)
            current_app.logger.info(prepared.body)
            current_app.logger.info(prepared.headers)
            return prepared

        return req



from dohq_artifactory import User
from artifactory import ArtifactoryPath

import inspect

def test(artifactory_generic_repository):
    # a = ArtifactoryPath(artifactory) #, auth=('admin', 'password'))

    raise Exception(inspect.getmembers(artifactory_generic_repository))



# -*- coding: utf-8 -*-
# (c) 2023 Brian Scholer (@briantist)

from datetime import datetime

from . import bp

@bp.route('/imports/collections/0')
@bp.route('/imports/collections/0/', endpoint='import_singleton')
def import_singleton():
    out = {
        'state': 'completed',
        'finished_at': datetime.utcnow(),
    }
    return out



import pytest
import requests
from urllib.parse import urljoin
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

from artifactory import ArtifactoryPath, RepositoryLocal
from dohq_artifactory import User


@pytest.fixture(autouse=True, scope='session')
def broken():
    pytest.skip('Integration tests are incomplete and disabled.')


@pytest.fixture(scope='session')
def artifactory(session_scoped_container_getter):
    request_session = requests.Session()
    retries = Retry(
        total=30,
        backoff_factor=0.1,
        status_forcelist=[500, 502, 503, 504],
    )
    request_session.mount('http://', HTTPAdapter(max_retries=retries))

    service = session_scoped_container_getter.get("artifactory").network_info[0]
    artifactory_url = "http://%s:%s/artifactory" % (service.hostname, service.host_port)
    readiness_url = "%s/api/v1/system/readiness" % (artifactory_url,)
    request_session.get(readiness_url)
    return ArtifactoryPath(artifactory_url)
    # return request_session, api_url


@pytest.fixture(scope='session')
def artifactory_user(artifactory):
    user = User(artifactory, name='admin', password='password')
    apikey = user.api_key.get()
    if not apikey:
        user.api_key.create()

    return user


@pytest.fixture(scope='session')
def artifactory_authed(artifactory, artifactory_user):
    return ArtifactoryPath(artifactory, apikey=artifactory_user.api_key.get())


@pytest.fixture(scope='session')
def artifactory_generic_repository(artifactory_authed):
    name = 'example-repo-local'
    # repo = artifactory_authed.find_repository_local(name)
    # ^ can't use this, requires pro license 🙄

    for repo in artifactory_authed.get_repositories():
        if repo.name == name:
            if repo.package_type == RepositoryLocal.GENERIC:
                return repo
            else:
                raise ValueError("existing repository is the wrong type.")

    # this is going to fail without a pro license
    repo = RepositoryLocal(artifactory_authed, name=name, package_type=RepositoryLocal.GENERIC)
    repo.create()

    return repo


@pytest.fixture
def repository(artifactory):
    return ArtifactoryPath(f"{artifactory}/repo/subpath")



# -*- coding: utf-8 -*-
# (c) 2023 Brian Scholer (@briantist)

from flask import Blueprint

bp = Blueprint('v3', __name__, url_prefix='/v3')

from . import (
    collection_imports,
    collections,
)



# -*- coding: utf-8 -*-
# (c) 2020 Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+
#     (see https://www.gnu.org/licenses/gpl-3.0.txt)

# this routine lovingly copied from:
# https://github.com/sivel/amanda/commit/ca3f59ffbe2801c230151aba342157b337c98a24

import sys
import tarfile


ENCODING = sys.getfilesystemencoding()


def iter_tar(f):
    while True:
        header = f.read(512)
        if not header:
            break

        typ = header[156:157]
        name = tarfile.nts(header[0:100], ENCODING, 'surrogateescape')
        prefix = tarfile.nts(header[345:500], ENCODING, 'surrogateescape')

        if prefix and typ not in tarfile.GNU_TYPES:
            name = f'{prefix}/{name}'

        if not name:
            continue

        size = tarfile.nti(header[124:136])
        if size:
            contents = tarfile.nts(f.read(size), ENCODING, 'surrogateescape')
        else:
            contents = None

        if name != '././@PaxHeader':
            yield name, contents

        mod = f.tell() % 512
        if mod:
            f.seek(512 - mod, 1)



# -*- coding: utf-8 -*-
# (c) 2023 Brian Scholer (@briantist)

import sys
import pytest

import semver
from datetime import datetime, timedelta, timezone
from pytest_mock import MockFixture

from artifactory import ArtifactoryPath

from galactory.models import CollectionData


@pytest.mark.parametrize('collection_info', [{}])
@pytest.mark.parametrize('created_datetime', [datetime.now(timezone.utc)])
@pytest.mark.parametrize('modified_datetime', [datetime.now(timezone.utc)])
@pytest.mark.parametrize('filename', ['fake-file'])
@pytest.mark.parametrize('mime_type', ['fake-mime'])
@pytest.mark.parametrize('name', ['collection_name'])
@pytest.mark.parametrize('namespace', ['a_namespace'])
@pytest.mark.parametrize('sha256', ['m-m-m-my-sha-'])
@pytest.mark.parametrize('size', [-1, 0, 1])
@pytest.mark.parametrize('version', ['invalid', '1.2.3', '1.2.3-dev0'])
def test_collectiondata_init(
    collection_info,
    created_datetime,
    modified_datetime,
    filename,
    mime_type,
    name,
    namespace,
    sha256,
    size,
    version,
):
    data = CollectionData(
        collection_info=collection_info,
        created_datetime=created_datetime,
        modified_datetime=modified_datetime,
        filename=filename,
        mime_type=mime_type,
        name=name,
        namespace=namespace,
        sha256=sha256,
        size=size,
        version=version,
    )

    assert isinstance(data, CollectionData)
    assert data.created_datetime.utcoffset() == timedelta(0)
    assert data.modified_datetime.utcoffset() == timedelta(0)
    assert datetime.fromisoformat(data.created) == data.created_datetime
    assert datetime.fromisoformat(data.modified) == data.modified_datetime
    assert data.fqcn == f"{data.namespace}.{data.name}"
    assert data.version == version
    try:
        ver = semver.VersionInfo.parse(version)
    except ValueError:
        with pytest.raises(ValueError, match=rf"^{version} is not valid SemVer string"):
            _ = data.semver
    else:
        assert isinstance(data.semver, semver.VersionInfo)
        assert semver.VersionInfo.parse(data.version) == data.semver
        assert data.semver == ver


_COLLECTION_INFO = [
    ('fake', 'one', '0.1.0', 'A'),
    ('fake', 'one', '0.1.0', 'B'),
    ('fake', 'one', '0.1.0-dev0', 'A'),
    ('fake', 'two', '2.0.0', 'C'),
    ('jake', 'two', '2.0.0', 'D'),
]


@pytest.mark.parametrize(
    ('namespace1', 'name1', 'version1', 'sha256_1'),
    _COLLECTION_INFO,
)
@pytest.mark.parametrize(
    ('namespace2', 'name2', 'version2', 'sha256_2'),
    _COLLECTION_INFO,
)
def test_collectiondata_compare_collections(
    namespace1, name1, version1, sha256_1,
    namespace2, name2, version2, sha256_2,
):
    _common = dict(
        collection_info={},
        created_datetime=datetime.now(timezone.utc),
        modified_datetime=datetime.now(timezone.utc),
        filename='fname',
        mime_type='mime',
        size=0,
    )

    colA = CollectionData(namespace=namespace1, name=name1, version=version1, sha256=sha256_1, **_common)
    colB = CollectionData(namespace=namespace2, name=name2, version=version2, sha256=sha256_2, **_common)

    verA = semver.VersionInfo.parse(version1)
    verB = semver.VersionInfo.parse(version2)

    assert (colA == colB) == (sha256_1 == sha256_2)

    if colA != colB:
        if name1 != name2 or namespace1 != namespace2:
            assert (colA < colB) == False
        elif verA.prerelease is None == verB.prerelease is None:
            assert (colA < colB) == (verA < verB)
        else:
            assert (colA < colB) == (verA.prerelease is not None)


@pytest.mark.parametrize(
    ('namespace1', 'name1', 'version1', 'sha256_1'),
    _COLLECTION_INFO,
)
@pytest.mark.parametrize(
    'other_version',
    list(map(lambda x: x[2], _COLLECTION_INFO)),
)
def test_collectiondata_compare_versioninfo(
    namespace1, name1, version1, sha256_1,
    other_version,
):
    _common = dict(
        collection_info={},
        created_datetime=datetime.now(timezone.utc),
        modified_datetime=datetime.now(timezone.utc),
        filename='fname',
        mime_type='mime',
        size=0,
    )

    col = CollectionData(namespace=namespace1, name=name1, version=version1, sha256=sha256_1, **_common)

    ver = semver.VersionInfo.parse(other_version)

    assert (col == ver) == (col.semver == ver)
    if col.is_prerelease == (ver.prerelease is not None):
        assert (col < ver) == (col.semver < ver)
    else:
        assert (col < ver) == (col.is_prerelease)


def test_collectiondata_compare_othertypes(mocker: MockFixture, collection_data: CollectionData):
    spy_eq = mocker.spy(CollectionData, '__eq__')
    spy_lt = mocker.spy(CollectionData, '__lt__')

    standin = 5

    assert collection_data != mocker.sentinel.whatever
    spy_eq.assert_called_once_with(collection_data, mocker.sentinel.whatever)
    assert spy_eq.spy_return is NotImplemented
    spy_eq.reset_mock()

    assert collection_data != standin
    spy_eq.assert_called_once_with(collection_data, standin)
    assert spy_eq.spy_return is NotImplemented
    spy_eq.reset_mock()

    assert standin != collection_data
    spy_eq.assert_called_once_with(collection_data, standin)
    assert spy_eq.spy_return is NotImplemented
    spy_eq.reset_mock()

    rmsg = r"not supported between instances of 'CollectionData' and"

    with pytest.raises(TypeError, match=rmsg):
        assert not (collection_data < standin)
    spy_lt.assert_called_once_with(collection_data, standin)
    assert spy_lt.spy_return is NotImplemented
    spy_lt.reset_mock()

    with pytest.raises(TypeError, match=rmsg):
        assert not (collection_data > standin)
    spy_lt.assert_called_once_with(collection_data, standin)
    assert spy_lt.spy_return is NotImplemented
    spy_lt.reset_mock()


@pytest.mark.xfail(condition=(sys.version_info >= (3, 10)), reason="Python>=3.10 tests don't work: https://github.com/briantist/galactory/issues/90")
def test_from_artifactory_repository(mocker: MockFixture, repository: ArtifactoryPath):
    spy = mocker.spy(CollectionData, '__init__')

    col = None
    for item in repository.iterdir():
        if item.is_dir():
            continue
        if item.name.endswith('.tar.gz'):
            props = item.properties
            stat = item.stat()
            col = CollectionData.from_artifactory_path(path=item, properties=props, stat=stat)
            spy.assert_called_once()
            spy.reset_mock()

    assert col is not None, "Error, no collections found."
